# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This file tells the default structure of the data
indexes = {
    "m00": 0,
    "m20": 1,
    "m22c": 2,
    "m22s": 3,
    "m40": 4,
    "m42c": 5,
    "m42s": 6,
    "v0": 7,
    "v1": 8,
    "v2": 9,
    "v3": 10,
    "v4": 11,
    "v5": 12,
    "v6": 13,
    "v7": 14,
    "v0_g1": 15,
    "v1_g1": 16,
    "v2_g1": 17,
    "v3_g1": 18,
    "v4_g1": 19,
    "v5_g1": 20,
    "v6_g1": 21,
    "v7_g1": 22,
    "v0_g2": 23,
    "v1_g2": 24,
    "v2_g2": 25,
    "v3_g2": 26,
    "v4_g2": 27,
    "v5_g2": 28,
    "v6_g2": 29,
    "v7_g2": 30,
}


col_names = [
    "fpfs_M00",
    "fpfs_M20",
    "fpfs_M22c",
    "fpfs_M22s",
    "fpfs_M40",
    "fpfs_M42c",
    "fpfs_M42s",
    "fpfs_v0",
    "fpfs_v1",
    "fpfs_v2",
    "fpfs_v3",
    "fpfs_v4",
    "fpfs_v5",
    "fpfs_v6",
    "fpfs_v7",
    "fpfs_v0r1",
    "fpfs_v1r1",
    "fpfs_v2r1",
    "fpfs_v3r1",
    "fpfs_v4r1",
    "fpfs_v5r1",
    "fpfs_v6r1",
    "fpfs_v7r1",
    "fpfs_v0r2",
    "fpfs_v1r2",
    "fpfs_v2r2",
    "fpfs_v3r2",
    "fpfs_v4r2",
    "fpfs_v5r2",
    "fpfs_v6r2",
    "fpfs_v7r2",
]

ncol = len(indexes)
npeak = 8



#!/usr/bin/env python
#
# FPFS shear estimator
# Copyright 20220312 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
import os
import gc
import jax
import fpfs
import pickle
import schwimmbad
import numpy as np
from scipy.optimize import minimize
from impt.fpfs.future import prepare_func_e
from impt.fpfs.default import indexes as did

import lsst.geom as lsstgeom
import lsst.afw.image as afwimage
from descwl_shear_sims.psfs import make_dm_psf

from argparse import ArgumentParser
from configparser import ConfigParser
import logging

logging.getLogger().setLevel(logging.CRITICAL)
logging.getLogger("jax").setLevel(logging.CRITICAL)


def get_processor_count(pool, args):
    if isinstance(pool, schwimmbad.MPIPool):
        # MPIPool
        from mpi4py import MPI

        return MPI.COMM_WORLD.Get_size() - 1
    elif isinstance(pool, schwimmbad.MultiPool):
        # MultiPool
        return args.n_cores
    else:
        # SerialPool
        return 1


band_map = {
    "g": 0,
    "r": 1,
    "i": 2,
    "z": 3,
    "a": 4,
}

nstd_map = {
    "g": 0.315,
    "r": 0.371,
    "i": 0.595,
    "z": 1.155,
    "a": 0.2793,
}

w_map = {
    "g": 0.12503653,
    "r": 0.47022727,
    "i": 0.30897575,
    "z": 0.09576044,
}


def get_seed_from_fname(fname, band):
    fid = int(fname.split("image-")[-1].split("_")[0]) + 212
    rid = int(fname.split("rot")[1][0])
    bid = band_map[band]
    return (fid * 2 + rid) * 4 + bid


class Worker(object):
    def __init__(
        self,
        config_name,
        min_id=0,
        max_id=1000,
        ncores=1,
        ratio=1.3,
        c0=4.0,
        c2=4.0,
        alpha=0.2,
        beta=0.8,
        sigma_mea=0.6,
        sigma_det=0.5,
    ):
        cparser = ConfigParser()
        cparser.read(config_name)

        # simulation parameter
        nids = max_id - min_id
        self.n_per_c = nids // ncores
        self.mid = nids % ncores
        self.min_id = min_id
        self.max_id = max_id
        self.rest_list = list(np.arange(ncores * self.n_per_c, nids) + min_id)
        print("number of files per core is: %d" % self.n_per_c)
        # print(self.min_id, self.max_id)

        self.img_dir = cparser.get("procsim", "img_dir")
        self.rcut = cparser.getint("FPFS", "rcut")
        self.nnord = cparser.getint("FPFS", "nnord", fallback=4)
        ngrid = 2 * self.rcut
        self.psf_fname = cparser.get("procsim", "psf_fname")
        self.band = cparser.get("survey", "band")
        self.nstd_f = nstd_map[self.band]
        self.noise_pow = np.ones((ngrid, ngrid)) * self.nstd_f**2.0 * ngrid**2.0
        self.scale = 0.2

        # setup processor
        self.ratio = ratio
        self.c0 = c0
        self.c2 = c2
        self.alpha = alpha
        self.beta = beta
        self.sigma_mea = sigma_mea
        self.sigma_det = sigma_det
        return

    def get_range(self, icore):
        ibeg = self.min_id + icore * self.n_per_c
        iend = min(ibeg + self.n_per_c, self.max_id)
        id_range = list(range(ibeg, iend))
        if icore < len(self.rest_list):
            id_range.append(self.rest_list[icore])
        return id_range

    def prepare_psf(self, exposure, rcut, ngrid2):
        # pad to (64, 64) and then cut off
        ngrid = 64
        beg = ngrid // 2 - rcut
        end = beg + 2 * rcut
        bbox = exposure.getBBox()
        bcent = bbox.getCenter()
        psf_model = exposure.getPsf()
        psf_array = psf_model.computeImage(lsstgeom.Point2I(bcent)).getArray()
        npad = (ngrid - psf_array.shape[0]) // 2
        psf_array2 = np.pad(psf_array, (npad + 1, npad), mode="constant")[
            beg:end, beg:end
        ]
        del npad
        # pad to exposure size
        npad = (ngrid2 - psf_array.shape[0]) // 2
        psf_array3 = np.pad(psf_array, (npad + 1, npad), mode="constant")
        return psf_array2, psf_array3

    def prepare_image(self, icore):
        fname = os.path.join(
            self.img_dir,
            "image-%05d_g1-0_rot0_%s.fits" % (icore, self.band),
        )
        exposure = afwimage.ExposureF.readFits(fname)
        # PSF
        with open(self.psf_fname, "rb") as f:
            psf_dict = pickle.load(f)
        psf_dm = make_dm_psf(**psf_dict)
        exposure.setPsf(psf_dm)
        del psf_dm
        self.image_nx = exposure.getWidth()
        psf_array2, psf_array3 = self.prepare_psf(
            exposure,
            self.rcut,
            self.image_nx,
        )
        # noise covariance
        noise_task = fpfs.image.measure_noise_cov(
            psf_array2,
            sigma_arcsec=self.sigma_mea,
            nnord=self.nnord,
            pix_scale=self.scale,
            sigma_detect=self.sigma_det,
        )
        cov_mat = np.array(noise_task.measure(self.noise_pow))
        del noise_task
        # image
        if False:
            seed = get_seed_from_fname(fname, self.band)
            rng = np.random.RandomState(seed)
            gal_array = exposure.getMaskedImage().getImage().getArray() + rng.normal(
                scale=self.nstd_f,
                size=(self.image_nx, self.image_nx),
            )
            del rng
        else:
            gal_array = exposure.getMaskedImage().getImage().getArray()
        # print(np.sqrt(np.diag(cov_mat)))
        del exposure
        gc.collect()
        return gal_array, psf_array2, psf_array3, cov_mat

    def process_image(self, icore):
        gal_array, psf_array2, psf_array3, cov_mat = self.prepare_image(icore)
        # measurement task
        meas_task = fpfs.image.measure_source(
            psf_array2,
            sigma_arcsec=self.sigma_mea,
            nnord=self.nnord,
            pix_scale=self.scale,
            sigma_detect=self.sigma_det,
        )

        std_modes = np.sqrt(np.diagonal(cov_mat))
        idm00 = fpfs.catalog.indexes["m00"]
        idv0 = fpfs.catalog.indexes["v0"]
        thres = 10.0 * std_modes[idm00] * self.scale**2.0
        thres2 = -1.5 * std_modes[idv0] * self.scale**2.0

        coords = fpfs.image.detect_sources(
            gal_array,
            psf_array3,
            sigmaf=meas_task.sigmaf,
            sigmaf_det=meas_task.sigmaf_det,
            thres=thres,
            thres2=thres2,
            bound=self.rcut + 5,
        )
        mm = meas_task.measure(gal_array, coords)
        del meas_task
        msk = (mm[:, did["m00"]] + mm[:, did["m20"]]) > 1e-5
        mm = mm[msk]
        del msk
        gc.collect()
        return mm, cov_mat

    def run(self, icore):
        id_range = self.get_range(icore)
        out = np.empty(len(id_range))
        # print("start core: %d, with id: %s" % (icore, id_range))
        for icount, ifield in enumerate(id_range):
            mm, cov_mat = self.process_image(ifield)
            e1, enoise, res1, rnoise = prepare_func_e(
                cov_mat=cov_mat,
                ratio=self.ratio,
                c0=self.c0,
                c2=self.c2,
                alpha=self.alpha,
                beta=self.beta,
                g_comp=1,
            )
            del cov_mat

            def fune(carry, ss):
                y = e1._obs_func(ss) - enoise._obs_func(ss)  # noqa
                return carry + y, None

            def funr(carry, ss):
                y = res1._obs_func(ss) - rnoise._obs_func(ss)  # noqa
                return carry + y, None

            e_sum, _ = jax.lax.scan(fune, 0.0, mm)
            r_sum, _ = jax.lax.scan(funr, 0.0, mm)
            out[icount] = float(e_sum / r_sum)
            del e1, enoise, res1, rnoise, mm
            gc.collect()
        return out


def process(args, pars):
    params = {
        "ratio": pars[0],
        "c0": pars[1],
        "c2": pars[2],
        "alpha": pars[3],
        "beta": pars[4],
        "sigma_mea": pars[5],
        "sigma_det": pars[6],
    }
    print("Current point: %s" % pars)
    with schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores) as pool:
        ncores = get_processor_count(pool, args)
        assert isinstance(ncores, int)
        core_list = np.arange(ncores)
        worker = Worker(
            args.config,
            min_id=args.min_id,
            max_id=args.max_id,
            ncores=ncores,
            **params,
        )
        outcome = np.hstack(list(pool.map(worker.run, core_list)))
        print(len(outcome))
        std = np.std(outcome)
        print("std: %s" % std)
    gc.collect()
    return std


if __name__ == "__main__":
    parser = ArgumentParser(description="fpfs procsim")
    parser.add_argument(
        "--config",
        required=True,
        type=str,
        help="configure file name",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--ncores",
        dest="n_cores",
        default=1,
        type=int,
        help="Number of processes (uses multiprocessing).",
    )
    parser.add_argument(
        "--min_id",
        default=0,
        type=int,
        help="id number, e.g. 0",
    )
    parser.add_argument(
        "--max_id",
        default=100,
        type=int,
        help="id number, e.g. 1000",
    )
    group.add_argument(
        "--mpi",
        dest="mpi",
        default=False,
        action="store_true",
        help="Run with MPI.",
    )
    args = parser.parse_args()
    process_opt = lambda _: process(args=args, pars=_)
    bounds = [
        (0.5, 2.5),
        (1.5, 40.0),
        (1.5, 40.0),
        (0.2, 1.2),
        (0.2, 1.2),
        (0.45, 0.75),
        (0.45, 0.75),
    ]
    x0 = np.array([1.8, 2.548, 22.774, 0.366, 0.960, 0.536, 0.507])
    op = {"maxiter": 400, "disp": False, "xtol": 1e-1}
    res = minimize(
        process_opt,
        x0,
        bounds=bounds,
        method="Nelder-Mead",
        # method="Powell",
    )
    print(res)



#!/usr/bin/env python
#
# FPFS shear estimator
# Copyright 20220312 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
import glob
import numpy as np
import astropy.io.fits as pyfits

from argparse import ArgumentParser
from configparser import ConfigParser


parser = ArgumentParser(description="fpfs procsim")
parser.add_argument(
    "--config",
    required=True,
    type=str,
    help="configure file name",
)

args = parser.parse_args()
cparser = ConfigParser()
cparser.read(args.config)
sum_dir = cparser.get("procsim", "sum_dir")
shear = cparser.getfloat("distortion", "shear_value")

flist = glob.glob("%s/bin_*.*.fits" % (sum_dir))
for fname in flist:
    mag = fname.split("/")[-1].split("bin_")[-1].split(".fits")[0]
    print("magnitude is: %s" % mag)
    a = pyfits.getdata(fname)
    a = a[np.argsort(a[:, 0])]
    nsim = a.shape[0]
    msk = np.isnan(a[:, 3])
    b = np.average(a, axis=0)
    c = np.std(a, axis=0)
    p = 99.7
    p2 = (100 - p) / 2.0
    p3 = p + p2
    print(b[1] / b[3] / shear / 2.0 - 1)
    print(
        (np.percentile(a[:, 1], p2) - np.average(a[:, 1]))
        / np.average(a[:, 3])
        / shear
        / 2.0
        / np.sqrt(nsim),
        (np.percentile(a[:, 1], p3) - np.average(a[:, 1]))
        / np.average(a[:, 3])
        / shear
        / 2.0
        / np.sqrt(nsim),
    )
    print(b[2] / b[3])
    print(
        (np.percentile(a[:, 2], p2) - np.average(a[:, 2]))
        / np.average(a[:, 3])
        / np.sqrt(nsim),
        (np.percentile(a[:, 1], p3) - np.average(a[:, 1]))
        / np.average(a[:, 3])
        / np.sqrt(nsim),
    )



#!/usr/bin/env python
#
# FPFS shear estimator
# Copyright 20220312 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
import os
import gc
import jax
import impt
import time
import fitsio
import schwimmbad
import numpy as np
from impt.fpfs.future import prepare_func_e

from argparse import ArgumentParser
from configparser import ConfigParser


def get_processor_count(pool, args):
    if isinstance(pool, schwimmbad.MPIPool):
        # MPIPool
        from mpi4py import MPI

        return MPI.COMM_WORLD.Get_size() - 1
    elif isinstance(pool, schwimmbad.MultiPool):
        # MultiPool
        return args.n_cores
    else:
        # SerialPool
        return 1


class Worker(object):
    def __init__(
        self,
        config_name,
        gver="g1",
        magcut=27.0,
        min_id=0,
        max_id=1000,
        ncores=1,
    ):
        cparser = ConfigParser()
        cparser.read(config_name)
        # survey parameter
        nids = max_id - min_id
        self.n_per_c = nids // ncores
        self.mid = nids % ncores
        self.min_id = min_id
        self.max_id = max_id
        self.rest_list = list(np.arange(ncores * self.n_per_c, nids) + min_id)
        print("number of files per core is: %d" % self.n_per_c)

        # setup processor
        self.catdir = cparser.get("procsim", "cat_dir")
        self.sum_dir = cparser.get("procsim", "sum_dir")
        ncov_fname = os.path.join(self.catdir, "cov_matrix.fits")
        self.cov_mat = fitsio.read(ncov_fname)
        # FPFS parameters
        self.ratio = cparser.getfloat("FPFS", "ratio")
        self.c0 = cparser.getfloat("FPFS", "c0")
        self.c2 = cparser.getfloat("FPFS", "c2")
        self.alpha = cparser.getfloat("FPFS", "alpha")
        self.beta = cparser.getfloat("FPFS", "beta")
        self.noise_rev = cparser.getboolean("FPFS", "noise_rev", fallback=True)
        # survey parameter
        self.magz = cparser.getfloat("survey", "mag_zero")
        # This task change the cut on one observable and see how the biases
        # changes.
        # Here is  the observable used for test
        self.upper_mag = magcut
        self.lower_m00 = 10 ** ((self.magz - self.upper_mag) / 2.5)
        # setup WL distortion parameter
        self.gver = gver
        self.ofname = os.path.join(
            self.sum_dir,
            "bin_%s.fits" % (self.upper_mag),
        )
        return

    def get_sum_e_r(self, in_nm, e1, enoise, res1, rnoise):
        assert os.path.isfile(
            in_nm
        ), "Cannot find input galaxy shear catalogs : %s " % (in_nm)
        mm = impt.fpfs.read_catalog(in_nm)
        # noise bias

        def fune(carry, ss):
            y = e1._obs_func(ss) - enoise._obs_func(ss)
            return carry + y, y

        def funr(carry, ss):
            y = res1._obs_func(ss) - rnoise._obs_func(ss)
            return carry + y, y

        e1_sum, _ = jax.lax.scan(fune, 0.0, mm)
        r1_sum, _ = jax.lax.scan(funr, 0.0, mm)
        del mm
        gc.collect()
        return e1_sum, r1_sum

    def get_range(self, icore):
        ibeg = self.min_id + icore * self.n_per_c
        iend = min(ibeg + self.n_per_c, self.max_id)
        id_range = list(range(ibeg, iend))
        if icore < len(self.rest_list):
            id_range.append(self.rest_list[icore])
        return id_range

    def run(self, icore):
        start_time = time.time()
        id_range = self.get_range(icore)
        out = np.zeros((len(id_range), 4))
        print("start core: %d, with id: %s" % (icore, id_range))
        for icount, ifield in enumerate(id_range):
            for irot in range(2):
                e1, enoise, res1, rnoise = prepare_func_e(
                    cov_mat=self.cov_mat,
                    snr_min=self.lower_m00 / np.sqrt(self.cov_mat[0, 0]),
                    ratio=self.ratio,
                    c0=self.c0,
                    c2=self.c2,
                    alpha=self.alpha,
                    beta=self.beta,
                    noise_rev=self.noise_rev,
                )
                b = "a"
                in_nm1 = os.path.join(
                    self.catdir,
                    "src-%05d_%s-0_rot%d_%s.fits" % (ifield, self.gver, irot, b),
                )
                e1_1, r1_1 = self.get_sum_e_r(in_nm1, e1, enoise, res1, rnoise)
                in_nm2 = os.path.join(
                    self.catdir,
                    "src-%05d_%s-1_rot%d_%s.fits" % (ifield, self.gver, irot, b),
                )
                e1_2, r1_2 = self.get_sum_e_r(in_nm2, e1, enoise, res1, rnoise)
                out[icount, 0] = ifield
                out[icount, 1] = out[icount, 1] + (e1_2 - e1_1)
                out[icount, 2] = out[icount, 2] + (e1_1 + e1_2) / 2.0
                out[icount, 3] = out[icount, 3] + (r1_1 + r1_2) / 2.0
                del e1, enoise, res1, rnoise
                jax.clear_backends()
                gc.collect()
        end_time = time.time()
        elapsed_time = (end_time - start_time) / 4.0
        print(f"Elapsed time: {elapsed_time} seconds")
        return out


if __name__ == "__main__":
    parser = ArgumentParser(description="fpfs procsim")
    parser.add_argument(
        "--runid",
        default=0,
        type=int,
        help="id number, e.g. 0",
    )
    parser.add_argument(
        "--min_id",
        default=0,
        type=int,
        help="id number, e.g. 0",
    )
    parser.add_argument(
        "--max_id",
        default=1000,
        type=int,
        help="id number, e.g. 1000",
    )
    parser.add_argument(
        "--magcut",
        default=27.0,
        type=float,
        help="id number, e.g. 0",
    )
    parser.add_argument(
        "--config",
        required=True,
        type=str,
        help="configure file name",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--ncores",
        dest="n_cores",
        default=1,
        type=int,
        help="Number of processes (uses multiprocessing).",
    )
    group.add_argument(
        "--mpi",
        dest="mpi",
        default=False,
        action="store_true",
        help="Run with MPI.",
    )
    args = parser.parse_args()
    cparser = ConfigParser()
    cparser.read(args.config)

    pool = schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores)
    ncores = get_processor_count(pool, args)
    assert isinstance(ncores, int)
    core_list = np.arange(ncores)
    worker = Worker(
        args.config,
        gver="g1",
        magcut=args.magcut,
        min_id=args.min_id,
        max_id=args.max_id,
        ncores=ncores,
    )
    summary_dirname = worker.sum_dir
    os.makedirs(summary_dirname, exist_ok=True)

    olist = pool.map(worker.run, core_list)
    pool.close()
    outcome = np.vstack(list(olist))
    fitsio.write(worker.ofname, outcome)



# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This file contains modules for perturbation functionals
# (vector perturbation: shear & tensor perturbation: noise)

# from jax import jit
# from functools import partial
import jax.numpy as jnp
from .base import NlBase

__all__ = ["RespG1", "RespG2", "BiasNoise"]


"""
For shear perturbation, we provide functionals to dervie the two shear response
functions.
This is a mapping from function to function.
"""


class RespG1(NlBase):
    """A Class to derive the shear response function [1st component] for an
    observable, following eq. (4) of
    https://arxiv.org/abs/2208.10522
    """

    def __init__(self, parent):
        """Initializes shear response object using a parent_obj object and
        a noise covariance matrix.
        """
        if not hasattr(parent, "_obs_grad_func"):
            raise TypeError("parent object does not has gradient operation")
        super().__init__(parent.params, parent, parent.lin_resp)
        return

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, x):
        """Returns the first-order shear response."""
        res = jnp.dot(
            self.parent._obs_grad_func(x),
            self.lin_resp._dg1(x),
        )
        return res


class RespG2(NlBase):
    """A Class to derive the shear response function [2nd component] for an
    observable, following eq. (4) of
    https://arxiv.org/abs/2208.10522
    """

    def __init__(self, parent):
        """Initializes shear response object using a parent_obj object and
        a noise covariance matrix.
        """
        if not hasattr(parent, "_obs_grad_func"):
            raise TypeError("parent object does not has gradient operation")
        super().__init__(parent.params, parent, parent.lin_resp)
        return

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, x):
        """Returns the first-order shear response."""
        res = jnp.dot(
            self.parent._obs_grad_func(x),
            self.lin_resp._dg2(x),
        )
        return res


"""
For noise perturbation, we provide a functional to dervie the correction function.
This is a mapping from function to function.
"""


class BiasNoise(NlBase):
    """A Class to derive the second-order noise perturbation function."""

    def __init__(self, parent, noise_cov):
        """Initializes noise response object using a parent_obj object and
        a noise covariance matrix.
        """
        if not hasattr(parent, "_obs_hessian_func"):
            raise TypeError("parent object does not has hessian operation")
        self.update_noise_cov(noise_cov)
        super().__init__(parent.params, parent, parent.lin_resp)
        return

    def update_noise_cov(self, noise_cov):
        self.noise_cov = noise_cov

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, x):
        """Returns the second-order noise response"""
        indexes = [[-2, -1], [-2, -1]]
        res = (
            jnp.tensordot(
                self.parent._obs_hessian_func(x),
                self.noise_cov,
                indexes,
            )
            / 2.0
        )
        return res


class BiasNoiseNull:
    """A Class disabling noise bias correction."""

    def __init__(self):
        """Initializes a null noise response object"""
        self._obs_func = self._base_func
        return

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, x):
        """Returns zero"""
        return 0.0



# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This file contains modules for nonlinear observables measured from images
# from jax import jit
# from functools import partial

import numpy as np
from flax import struct
from .default import npeak

from .default import indexes as did
from ..base import NlBase
from .linobs import FpfsLinResponse
from .utils import tsfunc2, smfunc, ssfunc2, ssfunc3
from ..perturb import RespG1, RespG2, BiasNoise, BiasNoiseNull

__all__ = [
    "FpfsExtParams",
    "FpfsExtE1",
    "FpfsExtE2",
]


class FpfsExtParams(struct.PyTreeNode):
    """FPFS parameter tree, these parameters are fixed in the tree"""

    # Exting parameter
    C0: float = struct.field(pytree_node=True, default=5.0)
    C2: float = struct.field(pytree_node=True, default=5.0)
    alpha: float = struct.field(pytree_node=True, default=1.3)
    beta: float = struct.field(pytree_node=True, default=1.2)

    # flux selection
    # cut on magntidue
    lower_m00: float = struct.field(pytree_node=False, default=0.2)
    # softening paramter for cut on flux
    sigma_m00: float = struct.field(pytree_node=False, default=0.2)

    # size selection
    # cut on size
    lower_r2: float = struct.field(pytree_node=False, default=0.03)
    upper_r2: float = struct.field(pytree_node=False, default=2.0)
    # softening paramter for cut on size
    sigma_r2: float = struct.field(pytree_node=False, default=0.2)

    # peak selection
    # cut on peak
    lower_v: float = struct.field(pytree_node=False, default=0.005)
    # softening parameter for cut on peak
    sigma_v: float = struct.field(pytree_node=False, default=0.2)

    sigma_min: float = struct.field(pytree_node=False, default=0.01)


class FpfsObsBase(NlBase):
    def __init__(self, params, parent=None, func_name="ts2"):
        if not isinstance(params, FpfsExtParams):
            raise TypeError("params is not FPFS parameters")
        lin_resp = FpfsLinResponse()
        if func_name == "sm":
            self.ufunc = smfunc
        elif func_name == "ts2":
            self.ufunc = tsfunc2
        elif func_name == "ss2":
            self.ufunc = ssfunc2
        elif func_name == "ss3":
            self.ufunc = ssfunc3
        else:
            raise ValueError("func_name: %s is not supported" % func_name)
        super().__init__(
            params=params,
            parent=parent,
            lin_resp=lin_resp,
        )


class FpfsExtE1(FpfsObsBase):
    """FPFS selection weight"""

    def __init__(self, params, parent=None, skip=1, func_name="ts2"):
        self.nmodes = 31
        self.skip = skip
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        # selection on flux
        w0 = self.ufunc(cat[did["m00"]], self.params.lower_m00, self.params.sigma_m00)
        # w2 = self.ufunc(
        #     cat[did["m00"]] + cat[did["m20"]],
        #     self.params.sigma_r2 * 3,
        #     self.params.sigma_r2,
        # )

        # selection on size (lower limit)
        # (M00 + M20) / M00 > lower_r2_lower
        r2l = cat[did["m00"]] * (1.0 - self.params.lower_r2) + cat[did["m20"]]
        w2l = self.ufunc(r2l, self.params.sigma_r2, self.params.sigma_r2)

        # selection on size (upper limit)
        # (M00 + M20) / M00 < upper_r2
        # M00 (1 - upper_r2) + M20 < 0
        # M00 (upper_r2 - 1) - M20 > 0
        r2u = cat[did["m00"]] * (self.params.upper_r2 - 1.0) - cat[did["m20"]]
        w2u = self.ufunc(r2u, self.params.sigma_r2, self.params.sigma_r2)
        wsel = w0 * w2l * w2u  # * w2

        # detection
        wdet = 1.0
        for i in range(0, npeak, self.skip):
            # v_i > lower_v
            wdet = wdet * self.ufunc(
                cat[did["v%d" % i]],
                self.params.lower_v,
                self.params.sigma_v,
            )

        # ellipticity
        denom = (cat[did["m00"]] + self.params.C0) ** self.params.alpha * (
            cat[did["m00"]] + cat[did["m20"]] + self.params.C2
        ) ** self.params.beta
        e1 = cat[did["m22c"]] / denom
        return wdet * wsel * e1


class FpfsExtE2(FpfsObsBase):
    """FPFS selection weight"""

    def __init__(self, params, parent=None, skip=1, func_name="ts2"):
        self.nmodes = 31
        self.skip = skip
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        # selection on flux
        w0 = self.ufunc(cat[did["m00"]], self.params.lower_m00, self.params.sigma_m00)
        # selection on size (lower limit)
        # (M00 + M20) / M00 > lower_r2_lower
        # M00 ( 1 - lower_r2_lower) + M20 > 0
        r2l = cat[did["m00"]] * (1.0 - self.params.lower_r2) + cat[did["m20"]]
        w2l = self.ufunc(r2l, self.params.sigma_r2, self.params.sigma_r2)
        # selection on size (upper limit)
        # (M00 + M20) / M00 < upper_r2
        # M00 (1 - upper_r2) + M20 < 0
        # M00 (upper_r2 - 1) - M20 > 0
        r2u = cat[did["m00"]] * (self.params.upper_r2 - 1.0) - cat[did["m20"]]
        w2u = self.ufunc(r2u, self.params.sigma_r2, self.params.sigma_r2)

        wsel = w0 * w2l * w2u
        # detection
        wdet = 1.0
        for i in range(0, npeak, self.skip):
            # v_i > lower_v
            wdet = wdet * self.ufunc(
                cat[did["v%d" % i]],
                self.params.lower_v,
                self.params.sigma_v,
            )

        # ellipticity
        denom = (cat[did["m00"]] + self.params.C0) ** self.params.alpha * (
            cat[did["m00"]] + cat[did["m20"]] + self.params.C2
        ) ** self.params.beta
        e2 = cat[did["m22s"]] / denom
        return wdet * wsel * e2


def prepare_func_e(
    cov_mat,
    ratio=1.81,
    c0=2.55,
    c2=25.6,
    alpha=0.27,
    beta=0.83,
    snr_min=12,
    r2_min=0.03,
    r2_max=2.0,
    g_comp=1,
    funcnm="ss2",
    noise_rev=True,
    sigma_min=0.01,
):
    if g_comp not in [1, 2]:
        raise ValueError("g_comp can only be 1 or 2")
    std_modes = np.sqrt(np.diagonal(cov_mat))
    std_m00 = std_modes[did["m00"]]
    std_m20 = np.sqrt(
        cov_mat[did["m00"], did["m00"]]
        + cov_mat[did["m20"], did["m20"]]
        + cov_mat[did["m00"], did["m20"]]
        + cov_mat[did["m20"], did["m00"]]
    )
    std_v0 = std_modes[did["v0"]]
    params = FpfsExtParams(
        C0=c0 * std_m00,
        C2=c2 * std_m20,
        alpha=alpha,
        beta=beta,
        lower_m00=snr_min * std_m00,
        lower_r2=r2_min,
        upper_r2=r2_max,
        lower_v=ratio * std_v0 * 0.4,
        sigma_m00=ratio * std_m00 + sigma_min,
        sigma_r2=ratio * std_m20 + sigma_min,
        sigma_v=ratio * std_v0 + sigma_min,
    )
    if g_comp == 1:
        ell = FpfsExtE1(params, func_name=funcnm)
        res = RespG1(ell)
    else:
        ell = FpfsExtE2(params, func_name=funcnm)
        res = RespG2(ell)
    if noise_rev:
        enoise = BiasNoise(ell, cov_mat)
        rnoise = BiasNoise(res, cov_mat)
    else:
        enoise = BiasNoiseNull()
        rnoise = BiasNoiseNull()
    return ell, enoise, res, rnoise



import numpy as np


def initialize_FPFS(fs, snlist, params):
    """Initialize FPFS catalog class
    Args:
        fs:                 fpfs.catalog.summary_stats
        snlist (ndarray):   ndarray of weight names
        params:             impt.fpfs params
    """
    cutsig = []
    cut = []
    for sn in snlist:
        if sn == "detect2":
            cutsig.append(params.sigma_v)
            cut.append(params.lower_v)
        elif sn == "M00":
            cutsig.append(params.sigma_m00)
            cut.append(params.lower_m00)
        elif sn == "R2":
            cutsig.append(params.sigma_r2)
            cut.append(params.lower_r2)
    cutsig = np.array(cutsig)
    cut = np.array(cut)
    fs.clear_outcomes()
    fs.update_selection_weight(snlist, cut, cutsig)
    fs.update_selection_bias(snlist, cut, cutsig)
    fs.update_ellsum()
    return fs



# impt autodiff pipline
# Copyright 20221114 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This is only a simple example of shear estimator
# You can define your own observable function
import numpy as np
from jax import jit
import jax.numpy as jnp
import jax


from .default import col_names, ncol

__all__ = ["fpfscov_to_imptcov", "ssfunc2", "ssfunc3", "tsfunc2", "smfunc"]


def fpfscov_to_imptcov(data):
    """Converts FPFS noise Covariance elements into a covariance matrix of
    lensPT.

    Args:
        data (ndarray):     input FPFS ellipticity catalog
    Returns:
        out (ndarray):      Covariance matrix
    """
    # the colum names
    # M00 -> N00; v1 -> V1
    ll = [cn[5:].replace("M", "N").replace("v", "V") for cn in col_names]
    out = np.zeros((ncol, ncol))
    for i in range(ncol):
        for j in range(ncol):
            try:
                try:
                    cname = "fpfs_%s%s" % (ll[i], ll[j])
                    out[i, j] = data[cname][0]
                except ValueError:
                    cname = "fpfs_%s%s" % (ll[j], ll[i])
                    out[i, j] = data[cname][0]
            except ValueError:
                out[i, j] = 0.0
    out = jnp.array(out)
    return out


@jit
def tsfunc2(x, mu, sigma):
    """Returns the C2 sinusoidal weight funciton
    This is for C2 sinusoidal based funciton

    Args:
        x (ndarray):    input data vector
        mu (float):     center of the cut
        sigma (float):  width of the selection function
    Returns:
        out (ndarray):  the weight funciton
    """
    t = (x - mu) / sigma

    def func(t):
        return 1.0 / 2.0 + t / 2.0 + 1.0 / 2.0 / jnp.pi * jnp.sin(t * jnp.pi)

    return jnp.piecewise(t, [t < -1, (t >= -1) & (t <= 1), t > 1], [0.0, func, 1.0])


@jit
def smfunc(x, mu, sigma):
    """Returns the sigmoid weight funciton

    Args:
        x (ndarray):    input data vector
        mu (float):     center of the cut
        sigma (float):  width of the selection function
    Returns:
        out (ndarray):  the weight funciton
    """
    # expx = jnp.exp(-(x - mu) / sigma)
    # # sigmoid function
    # return 1.0 / (1.0 + expx)

    t = (x - mu) / sigma
    # sigmoid function
    return jax.nn.sigmoid(t)


@jit
def ssfunc2(x, mu, sigma):
    """Returns the C3 smooth step weight funciton

    Args:
        x (ndarray):    input data vector
        mu (float):     center of the cut
        sigma (float):  width of the selection function
    Returns:
        out (ndarray):  the weight funciton
    """
    t = (x - mu) / sigma / 2.0 + 0.5

    def func(t):
        return 6 * t ** 5.0 - 15 * t ** 4.0 + 10 * t ** 3.0

    return jnp.piecewise(t, [t < 0, (t >= 0) & (t <= 1), t > 1], [0.0, func, 1.0])


@jit
def ssfunc3(x, mu, sigma):
    """Returns the C3 smooth step weight funciton

    Args:
        x (ndarray):    input data vector
        mu (float):     center of the cut
        sigma (float):  width of the selection function
    Returns:
        out (ndarray):  the weight funciton
    """
    t = (x - mu) / sigma / 2.0 + 0.5

    def func(t):
        return -20 * t ** 7 + 70 * t ** 6.0 - 84 * t ** 5.0 + 35 * t ** 4.0

    return jnp.piecewise(t, [t < 0, (t >= 0) & (t <= 1), t > 1], [0.0, func, 1.0])



# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This file contains pytrees for linear observables measured from images
# and functions to get their shear response

# from jax import jit
# from functools import partial
import jax.numpy as jnp
from fitsio import read as fitsread
import numpy.lib.recfunctions as rfn

from .default import indexes as did
from .default import col_names
from ..base import LinRespBase


__all__ = ["read_catalog", "FpfsLinResponse"]


"""
The following Classes are for FPFS. Feel free to extend the following system
or take it as an example to develop new system
"""
# TODO: Contact me if you are interested in adding or developing a new system
# of Observables


def read_catalog(fname):
    x = fitsread(fname)
    if x.dtype.names is not None:
        x = x[col_names]
        x = rfn.structured_to_unstructured(x, copy=False)
    return jnp.array(x, dtype=jnp.float64)


class FpfsLinResponse(LinRespBase):

    # @partial(jit, static_argnums=(0,))
    def _dg1(self, row):
        """Returns shear response array [first component] of shapelet pytree"""
        # shear response for shapelet modes
        fpfs_m00 = -jnp.sqrt(2.0) * row[did["m22c"]]
        fpfs_m20 = -jnp.sqrt(6.0) * row[did["m42c"]]
        fpfs_m22c = (row[did["m00"]] - row[did["m40"]]) / jnp.sqrt(2.0)
        # TODO: Include spin-4 term. Will add it when we have M44
        fpfs_m22s = jnp.zeros_like(row[did["m22s"]])
        # TODO: Incldue the shear response of M40 in the future. This is not
        # required in the FPFS shear estimation (v1~v3), so I set it to zero
        # here (But if you are interested in playing with shear response of
        # this term, please contact me.)
        fpfs_m40 = jnp.zeros_like(row[did["m40"]])
        fpfs_m42c = jnp.zeros_like(row[did["m42c"]])
        fpfs_m42s = jnp.zeros_like(row[did["m42s"]])
        out = jnp.stack(
            [
                fpfs_m00,
                fpfs_m20,
                fpfs_m22c,
                fpfs_m22s,
                fpfs_m40,
                fpfs_m42c,
                fpfs_m42s,
                row[did["v0_g1"]],
                row[did["v1_g1"]],
                row[did["v2_g1"]],
                row[did["v3_g1"]],
                row[did["v4_g1"]],
                row[did["v5_g1"]],
                row[did["v6_g1"]],
                row[did["v7_g1"]],
            ]
            + [0] * 16
        )
        return out

    # @partial(jit, static_argnums=(0,))
    def _dg2(self, row):
        """Returns shear response array [second component] of shapelet pytree"""
        fpfs_m00 = -jnp.sqrt(2.0) * row[did["m22s"]]
        fpfs_m20 = -jnp.sqrt(6.0) * row[did["m42s"]]
        # TODO: Include spin-4 term. Will add it when we have M44
        fpfs_m22c = jnp.zeros_like(row[did["m22c"]])
        fpfs_m22s = (row[did["m00"]] - row[did["m40"]]) / jnp.sqrt(2.0)
        # TODO: Incldue the shear response of M40 in the future. This is not
        # required in the FPFS shear estimation (v1~v3), so I set it to zero
        # here (But if you are interested in playing with shear response of
        # this term, please contact me.)
        fpfs_m40 = jnp.zeros_like(row[did["m40"]])
        fpfs_m42c = jnp.zeros_like(row[did["m42c"]])
        fpfs_m42s = jnp.zeros_like(row[did["m42s"]])
        out = jnp.stack(
            [
                fpfs_m00,
                fpfs_m20,
                fpfs_m22c,
                fpfs_m22s,
                fpfs_m40,
                fpfs_m42c,
                fpfs_m42s,
                row[did["v0_g2"]],
                row[did["v1_g2"]],
                row[did["v2_g2"]],
                row[did["v3_g2"]],
                row[did["v4_g2"]],
                row[did["v5_g2"]],
                row[did["v6_g2"]],
                row[did["v7_g2"]],
            ]
            + [0] * 16
        )
        return out



# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This file contains modules for nonlinear observables measured from images
# from jax import jit
# from functools import partial

from flax import struct
from .default import npeak

from .default import indexes as did
from ..base import NlBase
from .linobs import FpfsLinResponse
from .utils import tsfunc2, smfunc, ssfunc2, ssfunc3

__all__ = [
    "FpfsParams",
    "FpfsE1",
    "FpfsE2",
    "FpfsWeightSelect",
    "FpfsWeightDetect",
    "FpfsWeightE1",
    "FpfsWeightE2",
]

"""
The following Classes are for FPFS. Feel free to extend the following system
or take it as an example to develop new system
"""
# TODO: Contact me if you are interested in developing a new system of
# Observables


class FpfsParams(struct.PyTreeNode):
    """FPFS parameter tree, these parameters are fixed in the tree"""

    # Weighting parameter
    Const: float = struct.field(pytree_node=False, default=10.0)

    # flux selection
    # cut on magntidue
    lower_m00: float = struct.field(pytree_node=False, default=0.2)
    # softening paramter for cut on flux
    sigma_m00: float = struct.field(pytree_node=False, default=0.2)

    # size selection
    # cut on size
    lower_r2: float = struct.field(pytree_node=False, default=0.03)
    upper_r2: float = struct.field(pytree_node=False, default=2.0)
    # softening paramter for cut on size
    sigma_r2: float = struct.field(pytree_node=False, default=0.2)

    # peak selection
    # cut on peak
    lower_v: float = struct.field(pytree_node=False, default=0.005)
    # softening parameter for cut on peak
    sigma_v: float = struct.field(pytree_node=False, default=0.2)


class FpfsObsBase(NlBase):
    def __init__(self, params, parent=None, func_name="ts2"):
        if not isinstance(params, FpfsParams):
            raise TypeError("params is not FPFS parameters")
        lin_resp = FpfsLinResponse()
        if func_name == "sm":
            self.ufunc = smfunc
        elif func_name == "ts2":
            self.ufunc = tsfunc2
        elif func_name == "ss2":
            self.ufunc = ssfunc2
        elif func_name == "ss3":
            self.ufunc = ssfunc3
        else:
            raise ValueError("func_name: %s is not supported" % func_name)
        super().__init__(
            params=params,
            parent=parent,
            lin_resp=lin_resp,
        )


class FpfsWeightSelect(FpfsObsBase):
    """FPFS selection weight"""

    def __init__(self, params, parent=None, func_name="ts2"):
        self.nmodes = 31
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        # selection on flux
        w0 = self.ufunc(cat[did["m00"]], self.params.lower_m00, self.params.sigma_m00)

        # selection on size (lower limit)
        # (M00 + M20) / M00 > lower_r2_lower
        r2l = cat[did["m00"]] * (1.0 - self.params.lower_r2) + cat[did["m20"]]
        w2l = self.ufunc(r2l, self.params.sigma_r2, self.params.sigma_r2)

        # selection on size (upper limit)
        # (M00 + M20) / M00 < upper_r2
        r2u = cat[did["m00"]] * (self.params.upper_r2 - 1.0) - cat[did["m20"]]
        w2u = self.ufunc(r2u, self.params.sigma_r2, self.params.sigma_r2)
        # w2l = 1.
        # w2u = 1.
        out = w0 * w2l * w2u
        return out


class FpfsWeightDetect(FpfsObsBase):
    """FPFS detection weight"""

    def __init__(self, params, parent=None, skip=1, func_name="ts2"):
        self.nmodes = 31
        self.skip = skip
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        out = 1.0
        for i in range(0, npeak, self.skip):
            # v_i - M00 * lower_v > sigma_v
            vp = cat[did["v%d" % i]] - cat[did["m00"]] * self.params.lower_v
            out = out * self.ufunc(vp, self.params.sigma_v, self.params.sigma_v)
        return out


class FpfsE1(FpfsObsBase):
    """FPFS ellipticity (first component)"""

    def __init__(self, params, parent=None, func_name="ts2"):
        self.nmodes = 31
        super().__init__(
            params=params,
            parent=parent,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        return cat[did["m22c"]] / (cat[did["m00"]] + self.params.Const)


class FpfsE2(FpfsObsBase):
    """FPFS ellipticity (second component)"""

    def __init__(self, params, parent=None, func_name="ts2"):
        self.nmodes = 31
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        return cat[did["m22s"]] / (cat[did["m00"]] + self.params.Const)


class FpfsWeightE1(FpfsObsBase):
    """FPFS selection weight"""

    def __init__(self, params, parent=None, skip=1, func_name="ts2"):
        self.nmodes = 31
        self.skip = skip
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        # selection on flux
        w0 = self.ufunc(cat[did["m00"]], self.params.lower_m00, self.params.sigma_m00)

        # selection on size (lower limit)
        # (M00 + M20) / M00 > lower_r2_lower
        r2l = cat[did["m00"]] * (1.0 - self.params.lower_r2) + cat[did["m20"]]
        w2l = self.ufunc(r2l, self.params.sigma_r2, self.params.sigma_r2)

        # selection on size (upper limit)
        # (M00 + M20) / M00 < upper_r2
        # M00 ( 1 - lower_r2_lower) + M20 < 0
        r2u = cat[did["m00"]] * (self.params.upper_r2 - 1.0) - cat[did["m20"]]
        w2u = self.ufunc(r2u, self.params.sigma_r2, self.params.sigma_r2)
        wsel = w0 * w2l * w2u

        wdet = 1.0
        for i in range(0, npeak, self.skip):
            # v_i > lower_v
            wdet = wdet * self.ufunc(
                cat[did["v%d" % i]],
                self.params.lower_v,
                self.params.sigma_v,
            )

        e1 = cat[did["m22c"]] / (cat[did["m00"]] + self.params.Const)
        return wdet * wsel * e1


class FpfsWeightE2(FpfsObsBase):
    """FPFS selection weight"""

    def __init__(self, params, parent=None, skip=1, func_name="ts2"):
        self.nmodes = 31
        self.skip = skip
        super().__init__(
            params=params,
            parent=parent,
            func_name=func_name,
        )

    # @partial(jit, static_argnums=(0,))
    def _base_func(self, cat):
        # selection on flux
        w0 = self.ufunc(cat[did["m00"]], self.params.lower_m00, self.params.sigma_m00)

        # selection on size (lower limit)
        # (M00 + M20) / M00 > lower_r2_lower
        # M00 ( 1 - lower_r2_lower) + M20 > 0
        r2l = cat[did["m00"]] * (1.0 - self.params.lower_r2) + cat[did["m20"]]
        w2l = self.ufunc(r2l, self.params.sigma_r2, self.params.sigma_r2)

        # selection on size (upper limit)
        # (M00 + M20) / M00 < upper_r2
        r2u = cat[did["m00"]] * (self.params.upper_r2 - 1.0) - cat[did["m20"]]
        w2u = self.ufunc(r2u, 0.0, self.params.sigma_r2)
        wsel = w0 * w2l * w2u

        wdet = 1.0
        for i in range(0, npeak, self.skip):
            # v_i > lower_v
            wdet = wdet * self.ufunc(
                cat[did["v%d" % i]],
                self.params.lower_v,
                self.params.sigma_v,
            )
        e2 = cat[did["m22s"]] / (cat[did["m00"]] + self.params.Const)
        return wdet * wsel * e2



# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

import logging
from flax import struct
from jax import lax, grad, jacfwd, jacrev


logging.basicConfig(
    format="%(asctime)s %(message)s",
    datefmt="%Y/%m/%d %H:%M:%S --- ",
    level=logging.INFO,
)


class LinRespBase:
    def _dg1(*args):
        # each function has a basic function to apply to row and a function
        # wraped with lax
        raise NotImplementedError("You need to over-ride the _dg1 method")

    def _dg2(*args):
        # each function has a basic function to apply to row and a function
        # wraped with lax
        raise NotImplementedError("You need to over-ride the _dg2 method")

    def dg1(self, cat):
        return lax.map(self._dg1, cat)

    def dg2(self, cat):
        return lax.map(self._dg2, cat)


class NlBase:
    def __init__(self, params, parent=None, lin_resp=None):
        self.nmodes = 0
        if parent is not None:
            if not isinstance(parent, NlBase):
                raise ValueError("Input parent is not a instance of NlBase")
        self.parent = parent
        if lin_resp is not None:
            if not isinstance(lin_resp, LinRespBase):
                raise ValueError("Input lin_resp is not a instance of LinRespBase")
        self.lin_resp = lin_resp
        if not isinstance(params, struct.PyTreeNode):
            raise ValueError("Input parameter is not a instance of pyTreeNode")
        self.params = params
        self._set_obs_func(self._base_func)

    def _base_func(*args):
        raise NotImplementedError("You need to over-ride the _base_func method")

    def _set_obs_func(self, func):
        """Setup observable functions [func, derivative and Hessian]"""
        self._obs_func = func
        self._obs_grad_func = grad(self._obs_func, argnums=self.nmodes)
        self._obs_hessian_func = jacfwd(
            jacrev(
                self._obs_func,
            )
        )
        return

    def evaluate(self, cat):
        """Calls this observable function"""
        return lax.map(self._obs_func, cat)

    def grad(self, cat):
        """Calls the gradient vector function of observable function"""
        return lax.map(self._obs_grad_func, cat)

    def hessian(self, cat):
        """Calls the hessian matrix function of observable function"""
        return lax.map(self._obs_hessian_func, cat)

    def make_obs_new(self):
        out = NlBase(self.params, self, self.lin_resp)
        return out

    def __add__(self, other):
        obs = self.make_obs_new()
        if isinstance(other, NlBase):
            func = lambda x: self._obs_func(x) + other._obs_func(x)
        elif isinstance(other, (int, float)):
            func = lambda x: self._obs_func(x) + other
        else:
            raise TypeError("Cannot add %s to observable" % type(other))
        obs._set_obs_func(func)
        return obs

    def __sub__(self, other):
        obs = self.make_obs_new()
        if isinstance(other, NlBase):
            func = lambda x: self._obs_func(x) - other._obs_func(x)
        elif isinstance(other, (int, float)):
            func = lambda x: self._obs_func(x) - other
        else:
            raise TypeError("Cannot subtract %s to observable" % type(other))
        obs._set_obs_func(func)
        return obs

    def __mul__(self, other):
        obs = self.make_obs_new()
        if isinstance(other, NlBase):
            func = lambda x: self._obs_func(x) * other._obs_func(x)
        elif isinstance(other, (int, float)):
            func = lambda x: self._obs_func(x) * other
        else:
            raise TypeError("Cannot multiply %s to observable" % type(other))
        obs._set_obs_func(func)
        return obs

    def __truediv__(self, other):
        obs = self.make_obs_new()
        if isinstance(other, NlBase):
            func = lambda x: self._obs_func(x) / other._obs_func(x)
        elif isinstance(other, (int, float)):
            func = lambda x: self._obs_func(x) / other
        else:
            raise TypeError("Cannot multiply %s to observable" % type(other))
        obs._set_obs_func(func)
        return obs

    def __pow__(self, other):
        obs = self.make_obs_new()
        if isinstance(other, (int, float)):
            func = lambda x: self._obs_func(x) ** other
        else:
            raise TypeError("Cannot power %s to observable" % type(other))
        obs._set_obs_func(func)
        return obs



import os
from setuptools import setup, find_packages


this_dir = os.path.dirname(os.path.realpath(__file__))
__version__ = ""
fname = os.path.join(this_dir, "impt", "__version__.py")
with open(fname, "r") as ff:
    exec(ff.read())
long_description = open(os.path.join(this_dir, "README.md")).read()


scripts = [
    "bin/impt_config",
    "bin/impt_summary_hsc.py",
    "bin/impt_summary_desc_future.py",
    "bin/impt_std_future.py",
    "bin/impt_std_future_snr.py",
    "bin/impt_std_future_from_image.py",
    "bin/impt_make_plots.py",
]

setup(
    name="impt",
    version=__version__,
    description="Auto-diff Estimator of Lensing Perturbations",
    author="Xiangchong Li",
    author_email="mr.superonion@hotmail.com",
    python_requires=">=3.8",
    install_requires=[
        "numpy",
        "schwimmbad",
        "jax>=0.4.9",
        "jaxlib>=0.4.9",
        "flax",
        "fitsio",
    ],
    packages=find_packages(),
    scripts=scripts,
    include_package_data=True,
    zip_safe=False,
    url="https://github.com/mr-superonion/imPT/",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



#!/usr/bin/env python
#
# FPFS shear estimator
# Copyright 20220312 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
import os
import gc
import jax
import impt
import fitsio
import schwimmbad
import numpy as np
from impt.fpfs.future import prepare_func_e

from argparse import ArgumentParser
from configparser import ConfigParser
import logging

logging.getLogger().setLevel(logging.CRITICAL)
logging.getLogger("jax").setLevel(logging.CRITICAL)


def get_processor_count(pool, args):
    if isinstance(pool, schwimmbad.MPIPool):
        # MPIPool
        from mpi4py import MPI

        return MPI.COMM_WORLD.Get_size() - 1
    elif isinstance(pool, schwimmbad.MultiPool):
        # MultiPool
        return args.n_cores
    else:
        # SerialPool
        return 1


class Worker(object):
    def __init__(
        self,
        config_name,
        min_id=0,
        max_id=1000,
        ncores=1,
        ratio=1.3,
        c0=4.0,
        c2=4.0,
        alpha=0.2,
        beta=0.8,
        snr_min=12,
    ):
        cparser = ConfigParser()
        cparser.read(config_name)
        # simulation parameter
        nids = max_id - min_id
        self.n_per_c = nids // ncores
        self.mid = nids % ncores
        self.min_id = min_id
        self.max_id = max_id
        self.rest_list = list(np.arange(ncores * self.n_per_c, nids) + min_id)
        print("number of files per core is: %d" % self.n_per_c)

        # setup processor
        self.catdir = cparser.get("procsim", "cat_dir")
        ncov_fname = os.path.join(self.catdir, "cov_matrix.fits")
        self.cov_mat = fitsio.read(ncov_fname)
        self.ratio = ratio
        self.c0 = c0
        self.c2 = c2
        self.alpha = alpha
        self.beta = beta
        self.snr_min = snr_min
        return

    def get_range(self, icore):
        ibeg = self.min_id + icore * self.n_per_c
        iend = min(ibeg + self.n_per_c, self.max_id)
        id_range = list(range(ibeg, iend))
        if icore < len(self.rest_list):
            id_range.append(self.rest_list[icore])
        return id_range

    def get_sum_e_r(self, in_nm, e1, enoise, res1, rnoise):
        assert os.path.isfile(
            in_nm
        ), "Cannot find input galaxy shear catalogs : %s " % (in_nm)
        mm = impt.fpfs.read_catalog(in_nm)
        # noise bias

        def fune(carry, ss):
            y = e1._obs_func(ss) - enoise._obs_func(ss)
            return carry + y, y

        def funr(carry, ss):
            y = res1._obs_func(ss) - rnoise._obs_func(ss)
            return carry + y, y

        e1_sum, _ = jax.lax.scan(fune, 0.0, mm)
        r1_sum, _ = jax.lax.scan(funr, 0.0, mm)
        del mm
        gc.collect()
        return e1_sum, r1_sum

    def run(self, icore):
        id_range = self.get_range(icore)
        out = np.empty((len(id_range), 2))
        # print("start core: %d, with id: %s" % (icore, id_range))
        for icount, ifield in enumerate(id_range):
            e1, enoise, res1, rnoise = prepare_func_e(
                cov_mat=self.cov_mat,
                ratio=self.ratio,
                c0=self.c0,
                c2=self.c2,
                alpha=self.alpha,
                beta=self.beta,
                snr_min=self.snr_min,
                g_comp=1,
            )
            in_nm1 = os.path.join(
                self.catdir,
                "src-%05d_g1-1_rot0.fits" % (ifield),
            )
            ell, e_r = self.get_sum_e_r(in_nm1, e1, enoise, res1, rnoise)
            out[icount, 0] = ell
            out[icount, 1] = e_r
            del e1, enoise, res1, rnoise, ell, e_r
            gc.collect()
        return out


def process(args, pars, snr_min):
    params = {
        "ratio": pars[0],
        "c0": pars[1],
        "c2": pars[2],
        "alpha": pars[3],
        "beta": pars[4],
        "snr_min": snr_min,
    }
    with schwimmbad.choose_pool(mpi=args.mpi, processes=args.n_cores) as pool:
        ncores = get_processor_count(pool, args)
        assert isinstance(ncores, int)
        core_list = np.arange(ncores)
        worker = Worker(
            args.config,
            min_id=args.min_id,
            max_id=args.max_id,
            ncores=ncores,
            **params,
        )
        outcome = np.vstack(list(pool.map(worker.run, core_list)))
        std = np.std(outcome[:, 0]) / np.average(outcome[:, 1])
        print("std: %s" % std)
    return std


if __name__ == "__main__":
    parser = ArgumentParser(description="fpfs procsim")
    parser.add_argument(
        "--config",
        required=True,
        type=str,
        help="configure file name",
    )
    parser.add_argument(
        "--min_id",
        default=0,
        type=int,
        help="id number, e.g. 0",
    )
    parser.add_argument(
        "--max_id",
        default=100,
        type=int,
        help="id number, e.g. 1000",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--ncores",
        dest="n_cores",
        default=1,
        type=int,
        help="Number of processes (uses multiprocessing).",
    )
    group.add_argument(
        "--mpi",
        dest="mpi",
        default=False,
        action="store_true",
        help="Run with MPI.",
    )
    args = parser.parse_args()

    cparser = ConfigParser()
    cparser.read(args.config)
    sum_dir = cparser.get("procsim", "sum_dir")
    # FPFS parameters
    ratio = cparser.getfloat("FPFS", "ratio")
    c0 = cparser.getfloat("FPFS", "c0")
    c2 = cparser.getfloat("FPFS", "c2")
    alpha = cparser.getfloat("FPFS", "alpha")
    beta = cparser.getfloat("FPFS", "beta")
    pp = np.array([ratio, c0, c2, alpha, beta])
    process_opt = lambda x: process(args=args, pars=pp, snr_min=x)
    # snr_list = np.array([10, 12, 14.4, 17.28])  # , 20.736, 24.883])
    snr_list = np.logspace(1.02, 1.5, 10)
    outcomes = np.stack(list(map(process_opt, snr_list)))
    outcomes = np.vstack([snr_list.T, outcomes])
    print(outcomes)
    ofname = os.path.join(sum_dir, "snrcuts.fits")
    fitsio.write(ofname, outcomes)



# impt autodiff pipline
# Copyright 20221222 Xiangchong Li.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# python lib

# This file contains pytrees for linear observables measured from images
# and functions to get their shear response

from functools import partial

import jax.numpy as jnp
import numpy.lib.recfunctions as rfn
from fitsio import read as fitsread
from jax import jit

from ..base import LinRespBase
from .default import col_names
from .default import indexes as did

__all__ = ["read_catalog", "FpfsLinResponse"]


"""
The following Classes are for FPFS. Feel free to extend the following system
or take it as an example to develop new system
"""
# TODO: Contact me if you are interested in adding or developing a new system
# of Observables


def read_catalog(fname):
    x = fitsread(fname)
    if x.dtype.names is not None:
        x = x[col_names]
        x = rfn.structured_to_unstructured(x, copy=False)
    return jnp.array(x, dtype=jnp.float64)


class FpfsLinResponse(LinRespBase):
    @partial(jit, static_argnums=(0,))
    def _dg1(self, row):
        """Returns shear response array [first component] of shapelet pytree"""
        # shear response for shapelet modes
        fpfs_m00 = -jnp.sqrt(2.0) * row[did["m22c"]]
        fpfs_m20 = -jnp.sqrt(6.0) * row[did["m42c"]]
        fpfs_m22c = (row[did["m00"]] - row[did["m40"]]) / jnp.sqrt(2.0)
        # TODO: Include spin-4 term. Will add it when we have M44
        fpfs_m22s = jnp.zeros_like(row[did["m22s"]])
        # TODO: Incldue the shear response of M40 in the future. This is not
        # required in the FPFS shear estimation (v1~v3), so I set it to zero
        # here (But if you are interested in playing with shear response of
        # this term, please contact me.)
        fpfs_m40 = jnp.zeros_like(row[did["m40"]])  # fix this to 0
        fpfs_m42c = jnp.sqrt(6.0) / 2 * (row[did["m20"]] - row[did["m60"]])
        fpfs_m42s = jnp.zeros_like(row[did["m42s"]])
        fpfs_m60 = jnp.zeros_like(row[did["m42c"]])
        out = jnp.stack(
            [
                fpfs_m00,
                fpfs_m20,
                fpfs_m22c,
                fpfs_m22s,
                fpfs_m40,
                fpfs_m42c,
                fpfs_m42s,
                fpfs_m60,
                row[did["v0_g1"]],
                row[did["v1_g1"]],
                row[did["v2_g1"]],
                row[did["v3_g1"]],
                row[did["v4_g1"]],
                row[did["v5_g1"]],
                row[did["v6_g1"]],
                row[did["v7_g1"]],
            ]
            + [0] * 16
        )
        return out

    @partial(jit, static_argnums=(0,))
    def _dg2(self, row):
        """Returns shear response array [second component] of shapelet pytree"""
        fpfs_m00 = -jnp.sqrt(2.0) * row[did["m22s"]]
        fpfs_m20 = -jnp.sqrt(6.0) * row[did["m42s"]]
        # TODO: Include spin-4 term. Will add it when we have M44
        fpfs_m22c = jnp.zeros_like(row[did["m22c"]])
        fpfs_m22s = (row[did["m00"]] - row[did["m40"]]) / jnp.sqrt(2.0)
        # TODO: Incldue the shear response of M40 in the future. This is not
        # required in the FPFS shear estimation (v1~v3), so I set it to zero
        # here (But if you are interested in playing with shear response of
        # this term, please contact me.)
        fpfs_m40 = jnp.zeros_like(row[did["m40"]])
        fpfs_m42c = jnp.zeros_like(row[did["m42c"]])
        fpfs_m42s = jnp.sqrt(6.0) / 2 * (row[did["m20"]] - row[did["m60"]])
        fpfs_m60 = jnp.zeros_like(row[did["m42c"]])
        out = jnp.stack(
            [
                fpfs_m00,
                fpfs_m20,
                fpfs_m22c,
                fpfs_m22s,
                fpfs_m40,
                fpfs_m42c,
                fpfs_m42s,
                fpfs_m60,
                row[did["v0_g2"]],
                row[did["v1_g2"]],
                row[did["v2_g2"]],
                row[did["v3_g2"]],
                row[did["v4_g2"]],
                row[did["v5_g2"]],
                row[did["v6_g2"]],
                row[did["v7_g2"]],
            ]
            + [0] * 16
        )
        return out



# impt autodiff pipline
# flake8: noqa
import os
from .linobs import *
from .nlobs import *
from . import utils
from . import test_utils
from . import future

__data_dir__ = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

__all__ = []
__all__ += linobs.__all__
__all__ += nlobs.__all__
__all__ += ["utils"]
__all__ += ["test_utils"]
__all__ += ["future"]



# impt autodiff pipline
# flake8: noqa
import os
from .__version__ import __version__
from .perturb import *
from . import fpfs
from . import fpfs4

# We need accuracy is below 1e-6
from jax import config

config.update("jax_enable_x64", True)

__all__ = ["fpfs", "fpfs4"]
__all__ += perturb.__all__
